-- Written and maintained by zyewchi@gmail.com

STATRACK.LazyError = -1

local VOI_VALUE = 1
local VOI_TIMESTAMP = 2

local VOI_ABSCOND_FACTOR = 0.5 -- Almost definitely, don't exceed 0.618. a>b, (a - (a-b)*0.618) > (b + (a-b)*0.618^2). Exact: x = sqrt(5)/2 - 1/2, the solution for x^2 + x = 1. (The 'squishing' of the two variables cannot exceed their difference)
local VOI_ABSCOND_FACTOR_SQUARED = VOI_ABSCOND_FACTOR^2
local VOI_SWEEPER_SPREAD_FACTOR = 0.2
local MINIMUM_DIFFERENCE_FACTOR = 0.002

STATRACK.STATRACKSpecifier = "|cff10D0A0[STATRACK]".."|r "
STATRACK.infoSpecifier = "|cff50A0FF<INFO>".."|r: "
STATRACK.debugSpecifier = "|cffF080F0<DEBUG>".."|r: "
STATRACK.errSpecifier = "|cffC01050<ERR>".."|r: "

local MSG_HELP = 'Usage:\n "/statrack show, /statrack hide, /statrack reset"'

local hexCharacterTable = {[10] = 'A', [11] = 'B', [12] = 'C', [13] = 'D', [14] = 'E', [15] = 'F'}

SLASH_STATRACK1 = "/STATRACK"

SlashCmdList["STATRACK"] = function(msg)
	local args = STATRACK.SplitString(msg, "%s")
	if #args < 1 then
		print(STATRACK.STATRACKSpecifier.."Invalid command. Showing help..\n"..MSG_HELP)
	end
	
	if args[1]:lower() == "show" then 
		STATRACK.Show()
	elseif args[1]:lower() == "hide" then
		STATRACK.Hide()
	elseif args[1]:lower() == "reset" then
		STATRACK_VARS[UnitGUID("player")] = nil
		STATRACK.Reload()
	end
end

function STATRACK.SetVar(key, variable, forPlayer)
	if STATRACK_VARS == nil then
		STATRACK_VARS = {}
	end
	local thisPlayerGuid = UnitGUID("player")
	if STATRACK_VARS[thisPlayerGuid] == nil then
		STATRACK_VARS[thisPlayerGuid] = {}
	end
	if forPlayer then
		STATRACK_VARS[thisPlayerGuid][key] = variable
	else
		STATRACK_VARS[key] = variable
	end
end

function STATRACK.GetVar(key, forPlayer)
	local thisPlayerGuid = UnitGUID("player")
	if forPlayer and type(STATRACK_VARS) == "table" and STATRACK_VARS[thisPlayerGuid] ~= nil then
		return STATRACK_VARS[thisPlayerGuid][key]
	else
		return STATRACK_VARS[key]
	end
end

local function CreateValueOfIntTable(size)
	local tblVOI = {}
	for i=1,size,1 do
		tblVOI[i] = {}
	end
	return tblVOI
end

function STATRACK.IntToHex255(num)
	local num = max(0, min(255, num))
	local high = floor(num / 16)
	local low = (num % 16)
	
	return GetHex(high)..GetHex(low)
end

function STATRACK.GetHex(num)
	assert(num < 16)
	if num >= 10 then
		return hexCharacterTable[num]
	end
	return num
end

function STATRACK.RemoveAndCollapse(tbl, index, size)
	if index > size then return tbl end
	
	while index < size do
		local thisEntry = tbl[index]
		local nextData = tbl[index+1]
		thisEntry[VOI_VALUE] = nextData[VOI_VALUE]
		thisEntry[VOI_TIMESTAMP] = nextData[VOI_TIMESTAMP]
		index = index + 1
	end
	
	return tbl
end

function STATRACK.InsertAndGrow(tbl, index, size, value, timestamp)
	if index > size then error("You didn't want the table that big") end
	
	--print("IAG("..AQ.PrintableTable(tbl)..", "..index..", "..size..", {"..value[1]..", "..value[2].."})")
	
	local prev = value
	local prevTimestamp = timestamp
	local curr
	local currTimestamp
	
	while index <= size do
		local thisEntry = tbl[index]
		curr = thisEntry[VOI_VALUE]
		currTimestamp = thisEntry[VOI_TIMESTAMP]
		thisEntry[VOI_VALUE] = prev
		thisEntry[VOI_TIMESTAMP] = prevTimestamp
		prev = curr
		prevTimestamp = currTimestamp
		index = index + 1
	end
end

function STATRACK.SplitString(str, splitter)
	if str == nil or splitter == nil then
		error({code=STATRACK.LazyError, func="SplitString(str = "..(str == nil and "nil" or str)..", splitter = "..(splitter == nil and "nil" or splitter)..")"})
	end
	local str_arr = {}
	local i = 1
	for sub in str:gmatch(string.format("[^%s]+", splitter)) do
		str_arr[i] = sub
		i = i + 1
	end
	return str_arr
end


-- Locals to avoid declarations - May not be needed, may be a barely measurable performance improvement.
local thisTable
local thisData
local size
local maxSize

local willAbscond
local minimum
local maximum
local n
local minimumDifference
local differenceToTotal

local average
local variance
local sameness
local halfVariance
local weightedVariance
local valueScore

local skewed
local sweeperN
local nextSweeperN

local prevInt
local nthInt

local diff
local secondAbscondDiff

--- Reorders the (potentially one-value-out-of-order (by one or more)) array of absconded 
--- values, and absconds the new value into the set via the two values > it and <= it. 
--- The numerically furthest value being raised or lowered (depending of if the value is 
--- above or below average) by difference of the new value and furthest value times
--- the VOI_ABSCOND_FACTOR, then takes that difference and raises/lowers the closer value 
--- by the same difference*VOI_ABSCOND_FACTOR^2.
---
--- Holds maximums and minimums for a long time, and "squishes" them as they become stale
-------- STATRACK.ValueOfIntAbscond(id, newInt[, maxSize])
function STATRACK.ValueOfIntAbscond(id, newInt, ...)
	-- Initialize valueOfIntTable on first use
	if STATRACK.valueOfIntTable == nil then
		STATRACK.valueOfIntTable = {}
	end
	-- Initialize a table for this ID on first use
	if STATRACK.valueOfIntTable[id] == nil then
		local sizeArg = ...
		assert(type(sizeArg) == "number")
		STATRACK.valueOfIntTable[id] = {}
		STATRACK.valueOfIntTable[id].maxSize = sizeArg
		STATRACK.valueOfIntTable[id].data = {}
		STATRACK.valueOfIntTable[id].total = 0
		STATRACK.valueOfIntTable[id].sweeperN = 1
	end
	
	thisTable = STATRACK.valueOfIntTable[id]
	thisData = thisTable.data
	size = #thisData
	maxSize = thisTable.maxSize
	
	tableFull = (thisData[maxSize] ~= nil)
	minimum = (thisData[size] ~= nil and thisData[size] or newInt)
	maximum = (thisData[1] ~= nil and thisData[1] or newInt)
	n = 1
	minimumDifference = (maximum-minimum)*MINIMUM_DIFFERENCE_FACTOR
	
	differenceToTotal = 0
	
	average = thisTable.total / size
	
	if size == 0 then -- Received first value
		thisData[1] = newInt
		differenceToTotal = newInt
		size = 1
	else -- Insert / Abscond full table. O(n)
		local prevNthIntValue = 0xFFFFFFFF -- Player stats must never exceed 4.2 billion here..
		
		while n <= size+1 do -- n == maxsize will catch. +1 is for finding the (#tbl+1)th nil value when not full.
			nthInt = thisData[n]
			
			if nthInt == nil then -- Insert the value at end of this less-than-maxSized table
				differenceToTotal = newInt
				table.insert(thisData, n, newInt) -- Value was lowest
				size = n
				break
			end
			
			-- Quick on-the-fly reorder during search for absconding index.
			if prevNthIntValue < nthInt then
				thisData[n] = prevNthIntValue
				thisData[n-1] = nthInt
				nthInt = prevNthIntValue -- Util.c: n++; continue; would be used. Flag for this seems overkill.
			end
			
			-- Abscond mostly into furthest of the current and previous value when finding a lower value
			if newInt >= nthInt then
				prevInt = thisData[n-1]
				local prevWasFurthest = prevInt ~= nil and (prevInt - newInt) > (newInt - nthInt)
				
				if tableFull then
					local currAvg = thisTable.total / maxSize
					local aboveAvg = newInt > currAvg
					local thirdMaxSize = maxSize/3
					--local abscondFactor = VOI_ABSCOND_FACTOR * (1 - ((thirdMaxSize-(n%thirdMaxSize))/(1+thirdMaxSize))) -- The multiplier against VOI_ABSCOND_FACTOR causes easing of the absconding the further from the middle of the table we get. -- Allows a more accurate historical score based on an average-heavy minimum and maximum. Huge, rare bursts will only slightly affect the next high scores, but if they occur repeatedly, the high-score requirement will increase.
					local abscondFactor = VOI_ABSCOND_FACTOR * (1 - (abs(sin(deg((n-4)/2.6525))))) -- Cold-zones absconding are centered ~ 1, 8, 16, 24. .'. values are more efficiently absconded into the hot-zones at 4, 12, 20. Due to this the table is probably best at displaying 4 speeds. TODO Speeds.
					--if id == "Haste" then print(abscondFactor, halfMaxSize, n, currAvg, id) end
					
					if prevWasFurthest then -- thisData[n-1] was further from newIntv
						diff = (newInt - prevInt)*abscondFactor -- A -ve number
						if diff < -minimumDifference then 
							secondAbscondDiff = -diff*abscondFactor -- A lower magnitude, +ve number
							thisData[n-1] = prevInt + diff -- Abscond the further value
							thisData[n] = nthInt + secondAbscondDiff
							differenceToTotal = diff + secondAbscondDiff -- Slightly abscond the closer value
						else -- Value was very similar (MINIMUM_DIFFERENCE_FACTOR) 	-- n will be [2..maxSize]
							if aboveAvg then
								if n >= maxSize then
									n = n - 1
									prevInt = thisData[n-1]
									nthInt = prevInt
								end
								differenceToTotal = newInt - nthInt + minimumDifference
								thisData[n] = newInt + minimumDifference
								secondAbscondDiff = (newInt - thisData[n+1])*abscondFactor + minimumDifference
								thisData[n+1] = thisData[n+1] + secondAbscondDiff
								differenceToTotal = differenceToTotal + secondAbscondDiff
							else
								differenceToTotal = newInt - nthInt - minimumDifference
								thisData[n] = newInt - minimumDifference
								secondAbscondDiff = (newInt - prevInt)*abscondFactor - minimumDifference
								thisData[n-1] = thisData[n-1] + secondAbscondDiff
								differenceToTotal = differenceToTotal + secondAbscondDiff
							end
						end
						--if id == "Crit" then print(n, prevWasFurthest, newInt.."-"..prevInt, thisData[n-1], differenceToTotal) end
					else -- thisData[n] was further from newInt 
						diff = (newInt - nthInt)*abscondFactor -- A +ve number
						if diff > minimumDifference then
							thisData[n] = nthInt + diff -- Abscond the further value
							differenceToTotal = diff
							if n > 1 then
								secondAbscondDiff = -diff*abscondFactor -- A lower magnitude, -ve number
								thisData[n-1] = prevInt + secondAbscondDiff  -- Slightly abscond the closer value
								differenceToTotal = differenceToTotal + secondAbscondDiff
							end
						else -- Value was very similar (MINIMUM_DIFFERENCE_FACTOR) 	-- n will be [1..maxSize]
							if aboveAvg then
								if n >= maxSize then
									n = n - 1
									prevInt = thisData[n-1]
									nthInt = prevInt
								end
								differenceToTotal = newInt - nthInt + minimumDifference
								thisData[n] = newInt + minimumDifference
								secondAbscondDiff = (newInt - thisData[n+1])*abscondFactor + minimumDifference
								thisData[n+1] = thisData[n+1] + secondAbscondDiff
								differenceToTotal = differenceToTotal + secondAbscondDiff
							else
								if n < 2 then
									n = n + 1
									prevInt = nthInt
									nthInt = thisData[n]
									-- TODO Reorder step here
								end
								differenceToTotal = newInt - nthInt - minimumDifference
								thisData[n] = newInt - minimumDifference
								secondAbscondDiff = (newInt - prevInt)*abscondFactor - minimumDifference
								thisData[n-1] = thisData[n-1] + secondAbscondDiff
								differenceToTotal = differenceToTotal + secondAbscondDiff
							end
						end
						--if id == "Crit" then print(n, prevWasFurthest, newInt.."-"..nthInt, thisData[n], differenceToTotal) end
					end
					
					-- Settles the table of minimums and maximums over time.
					nextSweeperN = max(2, (thisTable.sweeperN+1) % (size+1))
					sweeperN = nextSweeperN - 1
					prevInt = thisData[sweeperN]
					nthInt = thisData[nextSweeperN]
					diff = prevInt - nthInt
					squishInt = diff*VOI_SWEEPER_SPREAD_FACTOR
					if diff > minimumDifference then
						thisData[sweeperN] = prevInt - squishInt
						thisData[nextSweeperN] = nthInt + squishInt
					end
					thisTable.sweeperN = nextSweeperN
					--if id=="Haste" then print(sweeperN, squishInt, differenceToTotal) end
				else -- Insert, in order, into the less-than-maxSized table
					differenceToTotal = newInt
					table.insert(thisData, n, newInt)
					size = n
				end
				
				break
			end
			if n == maxSize then -- This value defined a lower minimum
				differenceToTotal = newInt - nthInt
				thisData[n] = newInt
				break
			end
			prevNthIntValue = thisData[n]
			n = n + 1
		end
	end
	
	thisTable.total = thisTable.total + differenceToTotal
	average = thisTable.total / size
	variance = maximum - minimum
	halfVariance = variance / 2
	sameness = max(newInt, maximum) / minimum
	weightedVariance = halfVariance + ((average-minimum) > halfVariance and maximum-average or average-minimum)
	valueScore = 0.5 + (variance > 0 and sameness > 1.05 and (0.5*(newInt - average)/(weightedVariance)) or -0.25)
if STATRACK.DEBUG and id == "Vers" then print(STATRACK.STATRACKSpecifier..STATRACK.debugSpecifier.."Average is "..average..". (0.5 + (("..(newInt - average)..")/("..weightedVariance..")) = "..valueScore..".") end
	
	return max(0.0, min(1.0, valueScore))
end

--- Early testing algorithm, unused. Reliable, but short-lived data time-slot, requires much larger table for accurate measures (but will bomb your stat rating if you are given unusually high buffs, until those data-timestamps are removed)
-------- STATRACK.ValueOfIntOverTime(id, newInt, timestamp[, maxSize])
function STATRACK.ValueOfIntOverTime(id, newInt, timestamp, ...)
	-- Initialize valueOfIntTable on first use
	if STATRACK.valueOfIntTable == nil then
		STATRACK.valueOfIntTable = {}
	end
	-- Initialize a table for this ID on first use
	if STATRACK.valueOfIntTable[id] == nil then
		local sizeArg = ...
		assert(type(sizeArg) == "number")
		STATRACK.valueOfIntTable[id] = {}
		STATRACK.valueOfIntTable[id].maxSize = sizeArg
		STATRACK.valueOfIntTable[id].oldestTimestamp = timestamp
		STATRACK.valueOfIntTable[id].data = CreateValueOfIntTable(sizeArg, true)
		STATRACK.valueOfIntTable[id].total = 0
	end
	local thisTable = STATRACK.valueOfIntTable[id]
	local thisData = thisTable.data
	local size = #thisData
	local maxSize = thisTable.maxSize
	
	local thisOldestTimestamp = thisTable.oldestTimestamp
	local nextOldestTimestamp = 0xFFFFFFFF -- Prohibitively large, breaks if accidently set to oldestTimestamp

	local removingOldest = (thisData[size][VOI_TIMESTAMP] ~= nil)
	local newIntInserted = false
	local minimum = newInt
	local maximum = newInt
	local n = 1
	
	if size == 0 then
		thisData[1][VOI_VALUE] = newInt
		thisData[1][VOI_TIMESTAMP] = timestamp
		size = 1
	else
		if removingOldest and thisData[size][VOI_TIMESTAMP] == thisOldestTimestamp then
			thisTable.total = thisTable.total - thisData[size][VOI_VALUE]
			STATRACK.RemoveAndCollapse(thisData, size, maxSize)
			removingOldest = false -- Stop checking timestamps, we have removed oldest
		end
		while n <= size do
			local nthData = thisData[n]
			
			local nthInt = nthData[VOI_VALUE]
			local nthTimestamp = nthData[VOI_TIMESTAMP]
			
			if nthInt == nil then -- Minimum or maximum could be innacurrate for the first tick with this.
				if not newIntInserted then
					STATRACK.InsertAndGrow(thisData, n, maxSize, newInt, timestamp)
					size = n
				else
					size = n-1
				end
				break
			end
			
			-- Remove oldest timestamped element from array once found
			if removingOldest and nthTimestamp == thisOldestTimestamp then
				thisTable.total = thisTable.total - nthInt
				STATRACK.RemoveAndCollapse(thisData, n, maxSize)
				removingOldest = false
				
				-- Set currently searched data to the element collapsed to this index.
				nthData = thisData[n]
				
				nthInt = nthData[VOI_VALUE]
				nthTimestamp = nthData[VOI_TIMESTAMP]
				
				if nthData == nil then 	-- Minimum or maximum could be innacurrate for the first tick with this.
					STATRACK.InsertAndGrow(thisData, n, maxSize, newInt, timestamp)
					size = n
					break
				end
			end
			-- Find nextOldestTimestamp
			if nthTimestamp < nextOldestTimestamp then
				nextOldestTimestamp = nthTimestamp
			end
				
			-- Find minimum and maximum
			if nthInt < minimum then
				minimum = nthInt
			elseif nthInt > maximum then
				maximum = nthInt
			end
			
			if not newIntInserted and (n == size or newInt <= nthInt) then
				STATRACK.InsertAndGrow(thisData, n, maxSize, newInt, timestamp)
				newIntInserted = true
			end
			
			n = n + 1
		end
		thisTable.oldestTimestamp = nextOldestTimestamp
	end
	
	thisTable.total = thisTable.total + newInt
	-- Estimating Standard Deviation (or something) for efficiency
	local average = thisTable.total / min(size, maxSize)
	local variance = maximum - minimum
	local halfVariance = variance / 2
	local weightedVariance = halfVariance + ((average-minimum) > halfVariance and average-minimum or maximum-average)
	local valueScore = 0.5 + (variance > 0 and (0.5*(newInt - average)/(weightedVariance)) or 0)
if STATRACK.DEBUG and id == "Vers" then print(STATRACK.STATRACKSpecifier..STATRACK.debugSpecifier.."Average is "..average..". (0.5 + (("..(newInt - average)..")/("..weightedVariance..")) = "..valueScore..".") end
	
	return max(0.0, min(1.0, valueScore))
end